# coding = utf-8
import base64
from binascii import hexlify
import json
from os import urandom
import random
import time
from time import time as timestamp
from typing import BinaryIO, Union
from uuid import UUID
from bs4 import BeautifulSoup
import requests
import secmail
import names
from proudcatowner.torpy.http.requests import TorRequests

from fancy_text import fancy
from .helpers import cc_headers, constants, failed, log, print_banner, save, get_signature
from . import entities, exceptions
from .device import get_device_id
from .headers import get_api_headers
from .web_socket import Callbacks, SocketHandler
from colored import attr, back, fore, style


MB = fore.MAGENTA_3A + back.BLACK      + style.BLINK
OB = fore.ORANGE_1   + back.BLACK      + style.BLINK
BO = fore.BLACK      + back.ORANGE_1   + style.BLINK
OM = fore.ORANGE_1   + back.MAGENTA_3A + style.BLINK
WB = fore.BLACK      + back.WHITE      + style.BLINK
reset = attr('reset')

debug = True

print_banner()
input ("Press enter to start >> ")

class Client(Callbacks, SocketHandler):
    def __init__(self, deviceId: str = None, proxies: dict = None, certificatePath=None, socket_trace=False, socketDebugging=False):
        self.api = "https://service.narvii.com/api/v1"
        self.authenticated = False
        self.configured = False

        self.device_id = deviceId if deviceId is not None else get_device_id()

        SocketHandler.__init__(self, self, socket_trace=socket_trace, debug=socketDebugging)
        Callbacks.__init__(self, self)
        self.proxies = proxies
        self.certificatePath = certificatePath

        self.json = None
        self.sid = None
        self.userId = None
        self.account: entities.UserProfile = entities.UserProfile(None)
        self.profile: entities.UserProfile = entities.UserProfile(None)

    def _get_json_object(self, url, key: str = None):
        response = requests.get(url, headers=self.parse_headers(), proxies=self.proxies, verify=self.certificatePath)
        json_object = json.loads(response.text)
        if response.status_code != 200:
            exceptions.CheckException(json_object)
            return None

        return json_object[key] if key is not None else json_object

    def _get_entity(self, url, cls, key: str = None):
        json_object = self._get_json_object(url, key=key)
        return json_object and cls(json_object).Self

    def _post_get_json(self, url, data=None, headers=None, success_code=200, json_data=None):
        if data is not None:
            data = json.dumps(data)
        elif json_data is not None:
            data = json_data

        if headers is None:
            headers = self.parse_headers(data=data)

        response = requests.post(url, data=data, headers=headers, proxies=self.proxies, verify=self.certificatePath)
        json_object = json.loads(response.text)
        print(OB + f"\n{json_object}", reset)
        if json_object['api:message'].startswith("Whoa there!"):
            print(OB + "\nRate limited, changing IP address", reset)
            while True:
                try:
                    with TorRequests() as client:
                        with client.get_session() as session:
                            response = session.post(url, data=data, headers=headers, proxies=self.proxies, verify=self.certificatePath)
                except Exception as e:
                    print(e)
                    break

        if json_object["api:statuscode"] == 270:
            response.status_code == success_code
            print(OB + json_object["url"], reset)
            input("Press enter to skip verification >>")

        elif json_object['api:message'].startswith("Invalid Request"):
            print(OB + "Api changed?", reset)

        elif json_object["api:message"] == "Action not allowed.":
            print(OB + "Change IP address", reset)

        if response.status_code != success_code:
            pass
            return None

        return json_object

    def _post(self, url, data=None, headers=None, success_code=200):
        return None if self._post_get_json(url, data, headers, success_code) is None else success_code

    def _delete(self, url):
        response = requests.delete(url, headers=self.parse_headers(), proxies=self.proxies, verify=self.certificatePath)
        if response.status_code != 200:
            exceptions.CheckException(json.loads(response.text))
            return None

        return response.status_code

    def parse_headers(self, data: str = None):
        return get_api_headers(data=data, sid=self.sid, device_id=self.device_id)

    def deviceGenerator(self) -> str:
        device = requests.get(f"http://forevercynical.com/generate/deviceid").text
        return device

    def join_chat(self, chatId: str):
        """
        Join an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/chat/thread/{chatId}/member/{self.userId}"
        return self._post(url)

    def leave_chat(self, chatId: str):
        """
        Leave an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/chat/thread/{chatId}/member/{self.userId}"
        return self._delete(url)

    def join_voice_chat(self, comId: str, chatId: str, joinType: int = 1):
        """
        Joins a Voice Chat

        **Parameters**
            - **comId** : ID of the Community
            - **chatId** : ID of the Chat
        """

        data = {
            "o": {
                "ndcId": int(comId),
                "threadId": chatId,
                "joinRole": joinType,
                "id": "2154531"  
            },
            "t": 112
        }
        data = json.dumps(data)
        self.send(data)

    def verify(self, email: str, password: str, deviceId: str):
        url = f"{self.api}/g/s/auth/login"
        data = {
            "email": email,
            "v": 2,
            "secret": f"0 {password}",
            "deviceID": deviceId,
            "clientType": 100,
            "action": "normal",
            "timestamp": int(timestamp() * 1000)
        }
        json_object = self._post_get_json(url, data)
        self.run_socket()
        if json_object is not None:
            self.authenticated = True
            self.json = json_object
            self.sid = self.json["sid"]
            self.userId = self.json["account"]["uid"]
            self.account: entities.UserProfile = entities.UserProfile(self.json["account"]).Self
            self.profile: entities.UserProfile = entities.UserProfile(self.json["userProfile"]).Self
            self.start()
            return 200

        return None

    def request_verify_code(self, email: str):
        time.sleep(2)
        url = f"{self.api}/g/s/auth/request-security-validation"
        data = {
            "identity": email,
            "type": 1,
            "deviceID": reset_deviceId
        }
        return self._post_get_json(url, data)

    def update_password(self, email: str, new_password: str, captcha: str, link: str):
        time.sleep(2)
        url = f"{self.api}/g/s/auth/reset-password"
        data = {
            "updateSecret": f"0 {new_password}",
            "emailValidationContext": {
                "data": {
                    "code": captcha
                },
                "type": 1,
                "identity": email,
                "level": 2,
                "deviceID": reset_deviceId
            },
            "phoneNumberValidationContext": None,
            "deviceID": reset_deviceId
        }

        response = json_object = self._post(url, data)
        if response is None:
            print(f"Projected captcha: {captcha}")
            print(f"\nIf captcha matches link: \n{link}, \nand still fails.. try again later")
            failed(email)
            input("\nPress Enter to continue >>")
        if response == 200:
            save(email, new_password, reset_deviceId)
            short_device = f"{reset_deviceId[:5]}...{reset_deviceId[-5:]}"
            short_email = f"{email[:5]}...{email[-5:]}"
            short_pass = f"{new_password[:5]}...{new_password[-5:]}"
            self.print_pass(short_email, short_pass, short_device)
        return json_object

    def check_mail(self, email: str, new_password: str):
        try:
            time.sleep(.7)
            mail = secmail.SecMail()
            mails = mail.get_messages(email=email).id
            log(f"Checking inbox: {email}", do_log = debug)
            themail1 = mails[0]
            inbox = mail.read_message(email, themail1).htmlBody
            self.solve_captcha(email, new_password=new_password, inbox=inbox)
        except IndexError:
            self.check_mail(email, new_password=new_password)

    def solve_captcha(self, email: str, new_password: str, inbox):
        soup = BeautifulSoup(inbox, "html.parser")
        captcha = soup.find_all("a")
        captcha = captcha[0].get("href")
        link = captcha
        captcha = constants(captcha)
        self.update_password(email=email, new_password=new_password, captcha = captcha, link = link)

    def request_pass_code(self, email: str, new_password: str, deviceId: str):
        try:
            global reset_deviceId
            reset_deviceId = deviceId
            url = f"{self.api}/g/s/auth/request-security-validation"
            data = {
                "identity": email,
                "type": 1,
                "deviceID": reset_deviceId,
                "level": 2,
                "purpose":  "reset-password"
            }
            self._post(url, data)
            log('\nPassword Request Sent..', do_log = debug)
            self.check_mail(email=email, new_password=new_password)
        except Exception as e:
            print(e)

    def print_pass(self, short_email: str, short_pass: str,  short_device: str):
        color = fore.ORANGE_1 + back.BLACK + style.BLINK
        
        print(MB + "┄"*57, (reset))
        print(color + f"\tSuccessfully Changed Password!")
        print(color + f"\tEmail: {short_email}")
        print(color + f"\tPassword: {short_pass}")
        print(color + f"\tDevice: {short_device}")
        print(MB + "┄"*57, (reset))

    def change_password(self, email: str = None, new_password: str = None, deviceId: str = None):
        time.sleep(1.5)
        self.request_pass_code(email, new_password, deviceId)

    def generate_account(self, nickname: str = None, email: str = None, password: str = None, deviceId: str = None):
        """
        Generate Accounts.

        **Parameters**
            - **password** : input password for the accounts.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        self.request_generate_code(nickname, email, password, deviceId)

    def generate_email(self):
        mail = secmail.SecMail()
        email = mail.generate_email()
        return email

    def request_generate_code(self, nickname: str, email: str, password: str, deviceId: str):
        try:
            deviceId = deviceId
            data = {
            "identity": email,
            "type": 1,
            "deviceID": deviceId
            }
            data = json.dumps(data)
            requests.post(f"{self.api}/g/s/auth/request-security-validation", headers=self.parse_headers(data=data), data=data, proxies=self.proxies, verify=self.certificatePath)
            print(f"\nGenerating account with {email}..")
            self.check_mail_00(nickname, email = email, password = password, deviceId = deviceId)
        except Exception as e:
            print(e)

    def check_mail_00(self, nickname: str, email: str, password: str, deviceId: str):
        try:
            time.sleep(7)
            mail = secmail.SecMail()
            mails = mail.get_messages(email = email).id
            print(f"\nChecking inbox: {email}")
            themail1 = mails[0]
            inbox = mail.read_message(email, themail1).htmlBody
            self.solve_captcha_00(nickname, email = email, password = password, deviceId = deviceId, inbox = inbox)
        except IndexError or inbox:
            self.check_mail_00(nickname, email = email, password = password, deviceId = deviceId, inbox = inbox)

    def solve_captcha_00(self, nickname, email: str, password: str, deviceId: str, inbox):
        soup = BeautifulSoup(inbox, "html.parser")
        url = soup.find_all("a")
        url = url[0].get("href")
        print(f'\n{url}')
        code = input("\nEnter code here >> ")
        self.register(nickname, email = email, password = password, deviceId = deviceId, code = code, url1 = url)


    def generate_nickname(self):
        fancy_fonts = [fancy.light, fancy.bold, fancy.box]
        fancy_font = random.choice(fancy_fonts)
        nickname = names.get_first_name() + "⁷⁷⁷⁷"
        return fancy_font(nickname)


    def register(self, nickname: str, email: str, password: str, deviceId: str, code: str, url1: str):
            url = f"{self.api}/g/s/auth/register"
            data = ({
            "secret": f"0 {password}",
            "deviceID": deviceId,
            "email": email,
            "clientType": 100,
            "nickname": nickname,
            "latitude": 0,
            "longitude": 0,
            "address": None,
            "clientCallbackURL": "narviiapp://relogin",
            "validationContext": {
                "data": {
                    "code": code
                },
                "type": 1,
                "identity": email
            },
            "type": 1,
            "identity": email,
            "timestamp": int(time.time() * 1000)
            })  
            data = json.dumps(data)
            signature = get_signature(data)
            headers = cc_headers(deviceId, signature)
            response = requests.post(url, data=data, headers=headers)
            json_object = json.loads(response.text)
            if json_object["api:statuscode"] == 110 or 270:
                print(OB + "\nIP address banned, rotating proxy..\nPlease wait!", reset)
                while True:
                    with TorRequests() as client:
                        with client.get_session() as session:
                            response = session.post(url, data=data, headers=headers)
                            json_object = json.loads(response.text)
                            if response.status_code == 200:
                                print(BO + f'\napi:message - {json_object["api:message"]} - account generated', reset)
                                break
            if json_object["api:statuscode"] == 0: pass

            return json_object

    def login(self, email: str, password: str):
        """
        Login into an account.

        **Parameters**
            - **email** : Email of the account.
            - **password** : Password of the account.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/auth/login"
        data = {
            "email": email,
            "v": 2,
            "secret": f"0 {password}",
            "deviceID": self.device_id,
            "clientType": 100,
            "action": "normal",
            "timestamp": int(timestamp() * 1000)
        }
        json_object = self._post_get_json(url, data)
        self.run_socket() # TODO: do we run this even if login unsuccessful?
        if json_object is not None:
            self.authenticated = True
            self.json = json_object
            self.sid = self.json["sid"]
            self.userId = self.json["account"]["uid"]
            self.account: entities.UserProfile = entities.UserProfile(self.json["account"]).Self
            self.profile: entities.UserProfile = entities.UserProfile(self.json["userProfile"]).Self
            self.start()
            return 200

        return None


    def logout(self):
        """
        Logout from an account.

        **Parameters**
            - No parameters required.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/auth/logout"
        data = {
            "deviceID": self.device_id,
            "clientType": 100,
            "timestamp": int(timestamp() * 1000)
        }
        json_object = self._post_get_json(url, data)
        if json_object is not None:
            self.authenticated = False
            self.json = None
            self.sid = None
            self.userId = None
            self.account: None
            self.profile: None
            self.close()
            return 200

        return None

    def handle_socket_message(self, data):
        return self.resolve(data)

    def get_eventlog(self):
        url = f"{self.api}/g/s/eventlog/profile?language=en"
        return self._get_json_object(url)

    def get_communities(self, start: int = 0, size: int = 25):
        """
        List of Communities the account is in.

        **Parameters**
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`Community List <proudcatowner.entities.CommunityList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if not self.authenticated:
            raise exceptions.NotLoggedIn()

        url = f"{self.api}/g/s/community/joined?v=1&start={start}&size={size}"
        return self._get_entity(url, entities.CommunityList, "communityList")

    def sub_clients_profile(self, start: int = 0, size: int = 25):
        if not self.authenticated:
            raise exceptions.NotLoggedIn()

        url = f"{self.api}/g/s/community/joined?v=1&start={start}&size={size}"
        return self._get_json_object(url, "userInfoInCommunities")

    def get_user_info(self, userId: str):
        """
        Information of an User.

        **Parameters**
            - **userId** : ID of the User.

        **Returns**
            - **Success** : :meth:`User Object <proudcatowner.entities.UserProfile>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/user-profile/{userId}", entities.UserProfile
        return self._get_entity(url, entities.UserProfile, "userProfile")


    def get_chat_threads(self, start: int = 0, size: int = 25):
        """
        List of Chats the account is in.

        **Parameters**
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`Chat List <proudcatowner.entities.ThreadList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/chat/thread?type=joined-me&start={start}&size={size}"
        return self._get_entity(url, entities.ThreadList, "threadList")

    def get_chat_thread(self, chatId: str):
        """
        Get the Chat Object from an Chat ID.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : :meth:`Chat Object <proudcatowner.entities.Thread>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/chat/thread/{chatId}"
        return self._get_entity(url, entities.Thread, "thread")

    def get_chat_users(self, chatId: str, start: int = 0, size: int = 25):
        url = f"{self.api}/g/s/chat/thread/{chatId}/member?start={start}&size={size}&type=default&cv=1.2"
        return self._get_entity(url, entities.UserProfileList, "memberList")


    def start_chat(self, userId: Union[str, list], message: str, title: str = None, content: str = None, isGlobal: bool = False, publishToGlobal: bool = False):
        """
        Start an Chat with an User or List of Users.

        **Parameters**
            - **userId** : ID of the User or List of User IDs.
            - **message** : Starting Message.
            - **title** : Title of Group Chat.
            - **content** : Content of Group Chat.
            - **isGlobal** : If Group Chat is Global.
            - **publishToGlobal** : If Group Chat should show in Global.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if isinstance(userId, str):
            userIds = [userId]
        elif isinstance(userId, list):
            userIds = userId
        else:
            raise exceptions.WrongType()

        url = f"{self.api}/g/s/chat/thread"
        data = {
            "title": title,
            "inviteeUids": userIds,
            "initialMessageContent": message,
            "content": content,
            "timestamp": int(timestamp() * 1000)
        }
        if isGlobal is True:
            data["type"] = 2
            data["eventSource"] = "GlobalComposeMenu"
        else:
            data["type"] = 0

        if publishToGlobal is True:
            data["publishToGlobal"] = 1
        else:
            data["publishToGlobal"] = 0

        return self._post(url, data)

    def invite_to_chat(self, userId: Union[str, list], chatId: str):
        """
        Invite a User or List of Users to a Chat.

        **Parameters**
            - **userId** : ID of the User or List of User IDs.
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if isinstance(userId, str):
            userIds = [userId]
        elif isinstance(userId, list):
            userIds = userId
        else:
            raise exceptions.WrongType

        url = f"{self.api}/g/s/chat/thread/{chatId}/member/invite"
        data = {
            "uids": userIds,
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def kick(self, userId: str, chatId: str, allowRejoin: bool = True):
        allowRejoin = int(allowRejoin)
        url = f"{self.api}/g/s/chat/thread/{chatId}/member/{userId}?allowRejoin={allowRejoin}"
        return self._delete(url)

    def get_chat_messages(self, chatId: str, size: int = 25, pageToken: str = None):
        """
        List of Messages from an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.
            - *size* : Size of the list.
            - *size* : Size of the list.
            - *pageToken* : Next Page Token.

        **Returns**
            - **Success** : :meth:`Message List <proudcatowner.entities.MessageList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if pageToken is not None:
            url = f"{self.api}/g/s/chat/thread/{chatId}/message?v=2&pagingType=t&pageToken={pageToken}&size={size}"
        else:
            url = f"{self.api}/g/s/chat/thread/{chatId}/message?v=2&pagingType=t&size={size}"

        return self._get_entity(url, entities.GetMessages)

    def get_message_info(self, chatId: str, messageId: str):
        """
        Information of an Message from an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.
            - **messageId** : ID of the Message.

        **Returns**
            - **Success** : :meth:`Message Object <proudcatowner.entities.Message>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/chat/thread/{chatId}/message/{messageId}"
        return self._get_entity(url, entities.Message, "message")

    def get_community_info(self, comId: str):
        """
        Information of an Community.

        **Parameters**
            - **comId** : ID of the Community.

        **Returns**
            - **Success** : :meth:`Community Object <proudcatowner.entities.Community>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s-x{comId}/community/info?withInfluencerList=1&withTopicList=true&influencerListOrderStrategy=fansCount"
        return self._get_entity(url, entities.Community, "community")

    def search_community(self, aminoId: str):
        """
        Search a Community byt its Amino ID.

        **Parameters**
            - **aminoId** : Amino ID of the Community.

        **Returns**
            - **Success** : :meth:`Community List <proudcatowner.entities.CommunityList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/search/amino-id-and-link?q={aminoId}"
        json_object = self._get_json_object(url, "resultList")
        if json_object:
            return entities.CommunityList([com["refObject"] for com in json_object]).Self
        else:
            raise exceptions.CommunityNotFound(aminoId)


    def get_user_following(self, userId: str, start: int = 0, size: int = 25):
        """
        List of Users that the User is Following.

        **Parameters**
            - **userId** : ID of the User.
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`User List <proudcatowner.entities.UserProfileList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/user-profile/{userId}/joined?start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileList, "userProfileList")

    def get_user_followers(self, userId: str, start: int = 0, size: int = 25):
        """
        List of Users that are Following the User.

        **Parameters**
            - **userId** : ID of the User.
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`User List <proudcatowner.entities.UserProfileList>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/user-profile/{userId}/member?start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileList, "userProfileList")

    def send_message(self, chatId: str, message: str = None, messageType: int = 0, file: BinaryIO = None, fileType: str = None, replyTo: str = None, mentionUserIds: list = None, stickerId: str = None, embedId: str = None, embedType: int = None, embedLink: str = None, embedTitle: str = None, embedContent: str = None, embedImage: BinaryIO = None):
        """
        Send a Message to a Chat.

        **Parameters**
            - **message** : Message to be sent
            - **chatId** : ID of the Chat.
            - **file** : File to be sent.
            - **fileType** : Type of the file.
                - ``audio``, ``image``, ``gif``
            - **messageType** : Type of the Message.
            - **mentionUserIds** : List of User IDS to mention. '@' needed in the Message.
            - **replyTo** : Message ID to reply to.
            - **stickerId** : Sticker ID to be sent.
            - **embedTitle** : Title of the Embed.
            - **embedContent** : Content of the Embed.
            - **embedLink** : Link of the Embed.
            - **embedImage** : Image of the Embed.
            - **embedId** : ID of the Embed.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if message is not None and file is None:
            message = message.replace("<$", "‎‏").replace("$>", "‬‭")

        mentions = []
        if mentionUserIds:
            for mention_uid in mentionUserIds:
                mentions.append({"uid": mention_uid})

        if embedImage:
            embedImage = [[100, self.upload_media(embedImage, "image"), None]]

        url = f"{self.api}/g/s/chat/thread/{chatId}/message"
        data = {
            "type": messageType,
            "content": message,
            "clientRefId": int(timestamp() / 10 % 1000000000),
            "attachedObject": {
                "objectId": embedId,
                "objectType": embedType,
                "link": embedLink,
                "title": embedTitle,
                "content": embedContent,
                "mediaList": embedImage
            },
            "extensions": {"mentionedArray": mentions},
            "timestamp": int(timestamp() * 1000)
        }

        if replyTo:
            data["replyMessageId"] = replyTo

        if stickerId:
            data["content"] = None
            data["stickerId"] = stickerId
            data["type"] = 3

        if file:
            data["content"] = None
            if fileType == "audio":
                data["type"] = 2
                data["mediaType"] = 110

            elif fileType == "image":
                data["mediaType"] = 100
                data["mediaUploadValueContentType"] = "image/jpg"
                data["mediaUhqEnabled"] = True

            elif fileType == "gif":
                data["mediaType"] = 100
                data["mediaUploadValueContentType"] = "image/gif"
                data["mediaUhqEnabled"] = True

            else:
                raise exceptions.SpecifyType

            data["mediaUploadValue"] = base64.b64encode(file.read()).decode()

        return self._post(url, data)


    def delete_message(self, chatId: str, messageId: str, asStaff: bool = False, reason: str = None):
        """
        Delete a Message from a Chat.

        **Parameters**
            - **messageId** : ID of the Message.
            - **chatId** : ID of the Chat.
            - **asStaff** : If execute as a Staff member (Leader or Curator).
            - **reason** : Reason of the action to show on the Moderation History.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if not asStaff:
            url = f"{self.api}/g/s/chat/thread/{chatId}/message/{messageId}"
            return self._delete(url)

        url = f"{self.api}/g/s/chat/thread/{chatId}/message/{messageId}/admin"
        data = {
            "adminOpName": 102,
            "adminOpNote": {"content": reason},
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def send_coins(self, coins: int, blogId: str = None, chatId: str = None, objectId: str = None, transactionId: str = None):
        url = None
        if transactionId is None: transactionId = str(UUID(hexlify(urandom(16)).decode('ascii')))

        data = {
            "coins": coins,
            "tippingContext": {"transactionId": transactionId},
            "timestamp": int(timestamp() * 1000)
        }

        if blogId is not None:
            url = f"{self.api}/g/s/blog/{blogId}/tipping"

        if chatId is not None:
            url = f"{self.api}/g/s/chat/thread/{chatId}/tipping"

        if objectId is not None:
            data["objectId"] = objectId
            data["objectType"] = 2
            url = f"{self.api}/g/s/tipping"

        if url is None:
            raise exceptions.SpecifyType()

        return self._post(url, data)

    def follow(self, userId: Union[str, list]):
        """
        Follow an User or Multiple Users.

        **Parameters**
            - **userId** : ID of the User or List of IDs of the Users.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        if isinstance(userId, str):
            url = f"{self.api}/g/s/user-profile/{userId}/member"
            return self._post(url)
        elif isinstance(userId, list):
            url = f"{self.api}/g/s/user-profile/{self.userId}/joined"
            data = {
                "targetUidList": userId,
                "timestamp": int(timestamp() * 1000)
            }
            return self._post(url, data)
        else:
            raise exceptions.WrongType

    def unfollow(self, userId: str):
        """
        Unfollow an User.

        **Parameters**
            - **userId** : ID of the User.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/user-profile/{userId}/member/{self.userId}"
        return self._delete(url)

    def join_community(self, comId: str, invitationId: str = None):
        """
        Join a Community.

        **Parameters**
            - **comId** : ID of the Community.
            - **invitationId** : ID of the Invitation Code.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/x{comId}/s/community/join"
        data = {"timestamp": int(timestamp() * 1000)}
        if invitationId:
            data["invitationId"] = invitationId

        return self._post(url, data)

    def edit_profile(self, nickname: str = None, content: str = None, icon: BinaryIO = None, backgroundColor: str = None, backgroundImage: str = None, defaultBubbleId: str = None):
        """
        Edit account's Profile.

        **Parameters**
            - **nickname** : Nickname of the Profile.
            - **content** : Biography of the Profile.
            - **icon** : Icon of the Profile.
            - **backgroundImage** : Url of the Background Picture of the Profile.
            - **backgroundColor** : Hexadecimal Background Color of the Profile.
            - **defaultBubbleId** : Chat bubble ID.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/user-profile/{self.userId}"
        data = {
            "address": None,
            "latitude": 0,
            "longitude": 0,
            "mediaList": None,
            "eventSource": "UserProfileView",
            "timestamp": int(timestamp() * 1000)
        }

        if nickname:
            data["nickname"] = nickname
        if icon:
            data["icon"] = self.upload_media(icon, "image")
        if content:
            data["content"] = content
        if backgroundColor:
            data["extensions"] = {"style": {"backgroundColor": backgroundColor}}
        if backgroundImage:
            data["extensions"] = {"style": {"backgroundMediaList": [[100, backgroundImage, None, None, None]]}}
        if defaultBubbleId:
            data["extensions"] = {"defaultBubbleId": defaultBubbleId}

        return self._post(url, data)

    def like_blog(self, blogId: Union[str, list] = None, wikiId: str = None):
        """
        Like a Blog, Multiple Blogs or a Wiki.

        **Parameters**
            - **blogId** : ID of the Blog or List of IDs of the Blogs. (for Blogs)
            - **wikiId** : ID of the Wiki. (for Wikis)

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        data = {
            "value": 4,
            "timestamp": int(timestamp() * 1000)
        }

        if blogId:
            if isinstance(blogId, str):
                url = f"{self.api}/g/s/blog/{blogId}/g-vote?cv=1.2"
                data["eventSource"] = "UserProfileView"
            elif isinstance(blogId, list):
                url = f"{self.api}/g/s/feed/g-vote"
                data["targetIdList"] = blogId
            else:
                raise exceptions.WrongType(type(blogId))
        elif wikiId:
            url = f"{self.api}/g/s/item/{wikiId}/g-vote?cv=1.2"
            data["eventSource"] = "PostDetailView"
        else:
            raise exceptions.SpecifyType()

        return self._post(url, data)

    def get_wallet_info(self):
        """
        Get Information about the account's Wallet.

        **Parameters**
            - No parameters required.

        **Returns**
            - **Success** : :meth:`Wallet Object <proudcatowner.entities.WalletInfo>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/wallet"
        return self._get_entity(url, entities.WalletInfo, "wallet")

    def get_wallet_history(self, start: int = 0, size: int = 25):
        """
        Get the Wallet's History Information.

        **Parameters**
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`Wallet Object <proudcatowner.entities.WalletInfo>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/wallet/coin/history?start={start}&size={size}"
        return self._get_entity(url, entities.WalletHistory, "coinHistoryList")


    def get_from_link(self, code: str):
        """
        Get the Object Information from the Amino URL Code.

        **Parameters**
            - **code** : Code from the Amino URL.
                - ``http://aminoapps.com/p/EXAMPLE``, the ``code`` is 'EXAMPLE'.

        **Returns**
            - **Success** : :meth:`From Code Object <proudcatowner.entities.FromLink>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        url = f"{self.api}/g/s/link-resolution?q={code}"
        return self._get_entity(url, entities.FromLink, "linkInfoV2")

    def get_from_id(self, objectId: str, objectType: int, comId: str = None):
        """
        Get the Object Information from the Object ID and Type.

        **Parameters**
            - **objectID** : ID of the Object. User ID, Blog ID, etc.
            - **objectType** : Type of the Object.
            - *comId* : ID of the Community. Use if the Object is in a Community.

        **Returns**
            - **Success** : :meth:`From Code Object <proudcatowner.entities.FromLink>`

            - **Fail** : :meth:`Exceptions <proudcatowner.exceptions>`
        """
        data = {
            "objectId": objectId,
            "targetCode": 1,
            "objectType": objectType,
            "timestamp": int(timestamp() * 1000)
        }

        if comId:
            url = f"{self.api}/g/s-x{comId}/link-resolution"
        else:
            url = f"{self.api}/g/s/link-resolution"

        json_object = self._post(url, data)
        return json_object and entities.FromLink(json_object["linkInfoV2"]).Self


