from .device import get_device_id, user_agent
from .helpers import get_signature


def get_headers(data=None, sid=None, type=None, device_id: str = None, sig: str = None):
    if device_id is None:
        device_id = get_device_id()

    headers = {"NDCDEVICEID": device_id}
    if data:
        headers["Content-Length"] = str(len(data))
    if sid:
        headers["NDCAUTH"] = f"sid={sid}"
    if type:
        headers["Content-Type"] = type

    #TODO: sig is unused. We probably need the following:
    #if sig:
    #   headers["Whatever goes here"] = sig

    return headers


def get_api_headers(data=None, sid=None, type=None, device_id: str = None, sig: str = None):
    if device_id is None:
        device_id = get_device_id()

    headers = {
        "NDCDEVICEID": device_id,
        "Accept-Language": "en-US",
        "Content-Type": "text/javascript; charset=UTF-8",
        "User-Agent": user_agent(),
        "Host": "service.narvii.com",
        "Accept-Encoding": "gzip",
        "Connection": "Upgrade"
    }
    if data:
        headers["Content-Length"] = str(len(data))
        headers["NDC-MSG-SIG"] = get_signature(data)
    if sid:
        headers["NDCAUTH"] = f"sid={sid}"
    if type:
        headers["Content-Type"] = type
    if sig:
        headers["NDC-MSG-SIG"] = sig

    return headers


def get_vc_headers(data=None):
    vc_headers = {
        "Accept-Language": "en-US",
        "Content-Type": "application/json",
        "User-Agent": "Amino/45725 CFNetwork/1126 Darwin/19.5.0",  # Closest server (this one for me)
        "Host": "rt.applovin.com",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "Keep-Alive",
        "Accept": "*/*"
    }
    if data:
        vc_headers["Content-Length"] = str(len(data))

    return vc_headers
